package eu.jonahbauer.json.query.parser.ast;

import eu.jonahbauer.json.JsonValue;
import eu.jonahbauer.json.query.impl.Generator;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

public record JQReduceExpression(
        @NotNull JQExpression expression,
        @NotNull JQPatterns patterns,
        @NotNull JQExpression init,
        @NotNull JQExpression update
) implements JQExpression {

    @Override
    public @NotNull Generator<@Nullable JsonValue> evaluate(@NotNull Context context) {
        return init.evaluate(context).flatMap(initial -> {
            var state = new Object() {
                private JsonValue state = initial;
            };
            return Generator.concat(
                    expression.evaluate(context).flatMap(value -> patterns.bind(context, value, ctx -> {
                        var gen = update.evaluate(ctx.withRoot(state.state));
                        while (gen.hasNext()) {
                            state.state = gen.next();
                        }
                        return Generator.empty();
                    })),
                    Generator.of(() -> state.state)
            );
        });
    }

    @Override
    public boolean isConstant() {
        return expression.isConstant() && init.isConstant() && update.isConstant();
    }

    @Override
    public @NotNull String toString() {
        return "reduce " + expression + " as " + patterns + " (" + init + "; " + update + ")";
    }
}
